import { environment } from 'src/environments/environment';
import { HttpClient, HttpErrorResponse } from '@angular/common/http';
import { Observable, map, catchError, throwError } from 'rxjs';
import { ApiModel, PayloadModel, Empty, Area, AgentModel, JobModel, ResultModel, Job, NewPayloadModel, AssignedJobByIdModel } from '../models';
import { Injectable, Inject } from '@angular/core';
import { ErrorService } from './error.service';

type Status = "ok" | "err";

interface ServerResponse<T extends ApiModel> {
  status: Status,
  data: T | string
}

@Injectable({
  providedIn: 'root'
})
export class ApiTableService {

  constructor(
    private http: HttpClient,
    private errorService: ErrorService
  ) {
  }

  requestUrl = `${environment.server}/cmd/`;

  req<R extends ApiModel>(cmd: string): Observable<ServerResponse<R>> {
    return this.http.post<ServerResponse<R>>(this.requestUrl, cmd)
  }

  getOne<T extends ApiModel>(id: string, area: Area, brief: 'yes' | 'no' | 'auto' | null = null): Observable<T> {
    const request = `${area} read ${id}` + (brief !== null ? `-b=${brief}` : '')
    const resp = this.req<T[]>(request).pipe(
      map(resp => {
        if (resp.data.length === 0) {
          return {
            status: 'err' as Status,
            data: `${id} not found in ${area}`
          }
        }
        return {
          status: resp.status,
          data: resp.data[0]
        }
      }));
    return this.filterErrStatus(resp)
  }

  getAgent(id: string): Observable<AgentModel> {
    return this.getOne(id, 'agents')
  }

  getJob(id: string): Observable<Job> {
    return this.getOne(id, 'jobs')
  }

  getResult(id: string): Observable<ResultModel> {
    return this.getOne(id, 'map')
  }

  getPayload(id: string): Observable<PayloadModel> {
    return this.getOne(id, 'payloads')
  }

  getMany(area: Area): Observable<any[]> {
    return this.filterErrStatus(this.req(`${area} read`))
  }

  getAgents(): Observable<AgentModel[]> {
    return this.getMany('agents')
  }

  getJobs(): Observable<JobModel[]> {
    return this.getMany('jobs')
  }

  getResults(): Observable<ResultModel[]> {
    return this.getMany('map')
  }

  getPayloads(): Observable<PayloadModel[]> {
    return this.getMany('payloads')
  }

  update<T extends ApiModel>(item: T, area: Area): Observable<Empty> {
    return this.filterErrStatus(this.req(`${area} update '${JSON.stringify(item)}'`))
  }

  updateAgent(item: AgentModel): Observable<Empty> {
    return this.update(item, 'agents')
  }

  updateJob(item: JobModel): Observable<Empty> {
    return this.update(item, 'jobs')
  }

  updateResult(item: ResultModel): Observable<Empty> {
    return this.update(item, 'map')
  }

  updatePayload(item: PayloadModel): Observable<Empty> {
    return this.update(item, 'payloads')
  }

  delete(id: string, area: Area): Observable<Empty> {
    return this.filterErrStatus(this.req(`${area} delete ${id}`))
  }

  create<T extends ApiModel>(item: T | null, area: Area): Observable<string[]> {
    var serialized = '"{}"'
    if (item) {
      serialized = JSON.stringify(item);
    }
    return this.filterErrStatus(this.req(`${area} create '${serialized}'`))
  }

  createResult(item: AssignedJobByIdModel[]): Observable<string[]> {
    return this.create(item, 'map')
  }

  createPayload(item: NewPayloadModel): Observable<string[]> {
    return this.create(item, 'payloads')
  }

  filterErrStatus<R extends ApiModel>(obs: Observable<ServerResponse<R>>): Observable<R> {
    return obs.pipe(
      map(r => {
        if (r.status == 'err') {
          throw new Error(r.data as string)
        }
        return r.data as R
      }),
      catchError(this.errorHandler.bind(this)))
  }

  errorHandler(err: HttpErrorResponse, caught: any) {
    var error = err.error.data !== undefined ? JSON.stringify(err.error.data) : err.message;
    this.errorService.handle(error);
    return throwError(() => new Error());
  }
}