mod error;

use crate::{process_cmd, Args};
use actix_cors::Cors;
use actix_web::{get, middleware::Logger, post, web, App, HttpResponse, HttpServer, Responder};
use error::Error;
use futures_util::StreamExt;
use rust_embed::RustEmbed;
use std::borrow::Cow;
use structopt::StructOpt;
use u_lib::{api::HttpClient, unwrap_enum};

#[derive(RustEmbed)]
#[folder = "./src/gui/fe/dist/fe/"]
struct Files;

impl Files {
    pub fn get_static(path: impl AsRef<str>) -> Option<&'static [u8]> {
        let file = Self::get(path.as_ref())?.data;
        Some(unwrap_enum!(file, Cow::Borrowed))
    }
}

async fn spa_main() -> impl Responder {
    let index = Files::get_static("index.html").unwrap();
    HttpResponse::Ok().body(index)
}

#[get("/core/{path}")]
async fn resources_adapter(path: web::Path<(String,)>) -> impl Responder {
    let path = path.into_inner().0;
    let mimetype = mime_guess::from_path(&path).first_or_octet_stream();

    match Files::get_static(path) {
        Some(data) => HttpResponse::Ok()
            .content_type(mimetype.to_string())
            .body(data),
        None => HttpResponse::NotFound().finish(),
    }
}

#[post("/cmd/")]
async fn send_cmd(
    mut body: web::Payload,
    client: web::Data<HttpClient>,
) -> Result<impl Responder, Error> {
    let mut bytes = web::BytesMut::new();

    while let Some(item) = body.next().await {
        bytes.extend_from_slice(
            &item.map_err(|e| Error::JustError(format!("payload loading failure: {e}")))?,
        );
    }

    let cmd = String::from_utf8(bytes.to_vec())
        .map_err(|_| Error::JustError("cmd contains non-utf8 data".to_string()))?;
    let mut cmd = shlex::split(&cmd).ok_or(Error::JustError("argparse failed".to_string()))?;

    info!("cmd: {:?}", cmd);
    cmd.insert(0, String::from("u_panel"));

    let parsed_cmd = Args::from_iter_safe(cmd)?;
    let result = process_cmd(client.as_ref().clone(), parsed_cmd).await;
    let result_string = result.to_string();

    let response = if result.is_ok() {
        HttpResponse::Ok().body(result_string)
    } else {
        HttpResponse::BadRequest().body(result_string)
    };

    Ok(response)
}

pub async fn serve(client: HttpClient) -> anyhow::Result<()> {
    info!("Connecting to u_server...");
    client.ping().await?;

    let addr = "127.0.0.1:7799";
    info!("Connected, instanciating u_panel at http://{}", addr);

    HttpServer::new(move || {
        App::new()
            .wrap(Logger::default())
            .wrap(Cors::permissive())
            .app_data(web::Data::new(client.clone()))
            .service(send_cmd)
            .service(resources_adapter)
            .service(web::resource("/").to(spa_main))
            .service(web::resource("/{_}").to(spa_main))
    })
    .bind(addr)?
    .run()
    .await?;
    Ok(())
}
