use crate::error::Error;
use diesel::{pg::PgConnection, prelude::*, result::Error as DslError, Connection};
use std::collections::{HashMap, HashSet};
use std::mem::drop;
use u_lib::{
    db::PgAsyncPool,
    models::{schema, Agent, AssignedJob, AssignedJobById, Job, JobMeta, JobState, Payload},
    platform::Platform,
    types::Id,
};
use uuid::Uuid;

type Result<T> = std::result::Result<T, Error>;

pub struct PgRepo {
    pool: PgAsyncPool,
}

impl PgRepo {
    pub fn new(pool: PgAsyncPool) -> PgRepo {
        PgRepo { pool }
    }

    pub async fn interact<F, R>(&self, f: F) -> Result<R>
    where
        F: for<'c> FnOnce(UDB<'c>) -> Result<R>,
        F: Send + 'static,
        R: Send + 'static,
    {
        let connection = self.pool.get().await?;
        connection
            .interact(|conn| f(UDB { conn }))
            .await
            .expect("deadpool interaction failed")
    }

    pub async fn transaction<F, R>(&self, f: F) -> Result<R>
    where
        F: for<'c> FnOnce(UDB<'c>) -> Result<R>,
        F: Send + 'static,
        R: Send + 'static,
    {
        let conn = self.pool.get().await?;
        conn.interact(|c| c.transaction(|conn| f(UDB { conn })))
            .await
            .expect("deadpool interaction failed")
    }
}

pub struct UDB<'c> {
    conn: &'c mut PgConnection,
}

impl UDB<'_> {
    pub fn insert_jobs(&mut self, jobs: &[JobMeta]) -> Result<()> {
        use schema::jobs;

        diesel::insert_into(jobs::table)
            .values(jobs)
            .execute(self.conn)
            .map(drop)
            .map_err(with_err_ctx("Can't insert jobs"))
    }

    pub fn insert_payload(&mut self, payload: &Payload) -> Result<()> {
        use schema::payloads;

        diesel::insert_into(payloads::table)
            .values(payload)
            .execute(self.conn)
            .map(drop)
            .map_err(with_err_ctx(format!("Can't insert payload {payload:?}")))
    }

    pub fn get_job(&mut self, id: Id) -> Result<Option<Job>> {
        use schema::{jobs, payloads};

        let maybe_job_with_payload = jobs::table
            .left_join(payloads::table)
            .filter(jobs::id.eq(id))
            .first::<(JobMeta, Option<Payload>)>(self.conn)
            .optional()
            .map_err(with_err_ctx(format!("Can't get job {id}")))?;

        Ok(maybe_job_with_payload.map(|(job, payload)| Job { meta: job, payload }))
    }

    pub fn get_jobs(&mut self) -> Result<Vec<JobMeta>> {
        use schema::jobs;

        jobs::table
            .load(self.conn)
            .map_err(with_err_ctx("Can't get jobs"))
    }

    pub fn get_payload(&mut self, id: Id) -> Result<Option<Payload>> {
        use schema::payloads;

        payloads::table
            .filter(payloads::id.eq(id))
            .first(self.conn)
            .optional()
            .map_err(with_err_ctx(format!("Can't get payload {id}")))
    }

    pub fn get_payload_by_name(&mut self, name: String) -> Result<Option<Payload>> {
        use schema::payloads;

        payloads::table
            .filter(payloads::name.eq(&name))
            .first(self.conn)
            .optional()
            .map_err(with_err_ctx(format!("Can't get payload by name {name}")))
    }

    pub fn get_payloads(&mut self) -> Result<Vec<Payload>> {
        use schema::payloads;

        payloads::table
            .load(self.conn)
            .map_err(with_err_ctx("Can't get payloads"))
    }

    pub fn payload_exists(&mut self, payload_id: Id) -> Result<bool> {
        use schema::payloads;

        payloads::table
            .find(payload_id)
            .first::<Payload>(self.conn)
            .optional()
            .map(|r| r.is_some())
            .map_err(with_err_ctx("Can't check payload {payload_id}"))
    }

    pub fn get_job_by_alias(&mut self, alias: &str) -> Result<Option<Job>> {
        use schema::{jobs, payloads};

        let maybe_job_with_payload = jobs::table
            .left_join(payloads::table)
            .filter(jobs::alias.eq(alias))
            .first::<(JobMeta, Option<Payload>)>(self.conn)
            .optional()
            .map_err(with_err_ctx(format!("Can't get job by alias {alias}")))?;

        Ok(maybe_job_with_payload.map(|(job, payload_meta)| Job {
            meta: job,
            payload: payload_meta,
        }))
    }

    pub fn insert_result(&mut self, result: &AssignedJob) -> Result<()> {
        use schema::results;

        diesel::insert_into(results::table)
            .values(result)
            .execute(self.conn)
            .map_err(with_err_ctx(format!("Can't insert result {result:?}")))?;
        Ok(())
    }

    pub fn get_agent(&mut self, id: Id) -> Result<Option<Agent>> {
        use schema::agents;

        agents::table
            .filter(agents::id.eq(id))
            .first(self.conn)
            .optional()
            .map_err(with_err_ctx(format!("Can't get agent {id:?}")))
    }

    pub fn get_agents(&mut self) -> Result<Vec<Agent>> {
        use schema::agents;

        agents::table
            .load::<Agent>(self.conn)
            .map_err(with_err_ctx(format!("Can't get agents")))
    }

    pub fn update_job_status(&mut self, id: Id, status: JobState) -> Result<()> {
        use schema::results;

        diesel::update(results::table)
            .filter(results::id.eq(id))
            .set(results::state.eq(status))
            .execute(self.conn)
            .map_err(with_err_ctx(format!("Can't update status of job {id}")))?;
        Ok(())
    }

    //TODO: filters possibly could work in a wrong way, check
    pub fn get_assigned_jobs(
        &mut self,
        id: Option<Id>,
        personal: bool,
    ) -> Result<Vec<AssignedJob>> {
        use schema::results;

        let mut q = results::table.into_boxed();
        /*if id.is_some() {
            q = q.filter(results::agent_id.eq(id.unwrap()))
        }*/
        if personal {
            q = q.filter(
                results::state
                    .eq(JobState::Queued)
                    .and(results::agent_id.eq(id.unwrap())),
            )
        } else if id.is_some() {
            q = q
                .filter(results::agent_id.eq(id.unwrap()))
                .or_filter(results::job_id.eq(id.unwrap()))
                .or_filter(results::id.eq(id.unwrap()))
        }
        let result = q
            .load::<AssignedJob>(self.conn)
            .map_err(with_err_ctx("Can't get exact jobs"))?;
        Ok(result)
    }

    // todo: move to handlers
    pub fn assign_jobs(&mut self, assigned_jobs: &[AssignedJobById]) -> Result<()> {
        use schema::{jobs, results};

        struct JobBriefMeta {
            alias: Option<String>,
            target_platform: String,
        }

        let assigned_job_ids = HashSet::<Uuid>::from_iter(assigned_jobs.iter().map(|a| a.job_id));

        let jobs_meta = HashMap::<Id, JobBriefMeta>::from_iter(
            jobs::table
                .select((jobs::id, jobs::alias, jobs::target_platforms))
                .filter(jobs::id.eq_any(&assigned_job_ids))
                .load::<(Id, Option<String>, String)>(self.conn)
                .map_err(with_err_ctx(format!(
                    "Can't find jobs {:?}",
                    assigned_job_ids
                )))?
                .into_iter()
                .map(|(id, alias, target_platform)| {
                    (
                        id,
                        JobBriefMeta {
                            alias,
                            target_platform,
                        },
                    )
                }),
        );

        let existing_job_ids = HashSet::from_iter(jobs_meta.keys().copied());

        if assigned_job_ids != existing_job_ids {
            return Err(Error::ProcessingError(format!(
                "Jobs not found: {:?}",
                assigned_job_ids.difference(&existing_job_ids),
            )));
        }

        for ajob in assigned_jobs {
            let meta = &jobs_meta[&ajob.job_id];
            let agent_platform = match self.get_agent(ajob.agent_id)? {
                Some(agent) => Platform::new(&agent.platform),
                None => {
                    return Err(Error::ProcessingError(format!(
                        "Agent {} not found",
                        ajob.agent_id
                    )))
                }
            };
            if !agent_platform.matches(&meta.target_platform) {
                return Err(Error::InsuitablePlatform(
                    agent_platform.into_string(),
                    meta.target_platform.clone(),
                ));
            }
        }

        let job_requests = assigned_jobs
            .into_iter()
            .map(|a| AssignedJob {
                job_id: a.job_id,
                agent_id: a.agent_id,
                alias: jobs_meta[&a.job_id].alias.clone(),
                ..Default::default()
            })
            .collect::<Vec<AssignedJob>>();

        diesel::insert_into(results::table)
            .values(&job_requests)
            .execute(self.conn)
            .map(drop)
            .map_err(with_err_ctx("Can't assign jobs"))
    }

    pub fn del_jobs(&mut self, ids: &[Id]) -> Result<()> {
        use schema::jobs;

        diesel::delete(jobs::table)
            .filter(jobs::id.eq_any(ids))
            .execute(self.conn)
            .map(drop)
            .map_err(with_err_ctx("Can't delete jobs"))
    }

    pub fn del_results(&mut self, ids: &[Id]) -> Result<()> {
        use schema::results;

        diesel::delete(results::table)
            .filter(results::id.eq_any(ids))
            .execute(self.conn)
            .map(drop)
            .map_err(with_err_ctx("Can't delete results"))
    }

    pub fn del_agents(&mut self, ids: &[Id]) -> Result<()> {
        use schema::agents;

        diesel::delete(agents::table)
            .filter(agents::id.eq_any(ids))
            .execute(self.conn)
            .map(drop)
            .map_err(with_err_ctx("Can't delete agents"))
    }

    pub fn del_payloads(&mut self, ids: &[Id]) -> Result<()> {
        use schema::payloads;

        diesel::delete(payloads::table)
            .filter(payloads::id.eq_any(ids))
            .execute(self.conn)
            .map(drop)
            .map_err(with_err_ctx("Can't delete payloads"))
    }

    pub fn upsert_agent(&mut self, agent: &Agent) -> Result<()> {
        use schema::agents;

        diesel::insert_into(agents::table)
            .values(agent)
            .on_conflict(agents::id)
            .do_update()
            .set(agent)
            .execute(self.conn)
            .map_err(with_err_ctx(format!("Can't insert agent {agent:?}")))?;
        Ok(())
    }

    pub fn update_job(&mut self, job: &JobMeta) -> Result<()> {
        job.save_changes::<JobMeta>(self.conn)
            .map_err(with_err_ctx(format!("Can't update job {job:?}")))?;
        Ok(())
    }

    pub fn update_payload(&mut self, payload: &Payload) -> Result<()> {
        payload
            .save_changes::<Payload>(self.conn)
            .map_err(with_err_ctx(format!("Can't update payload {payload:?}")))?;
        Ok(())
    }

    pub fn update_result(&mut self, result: &AssignedJob) -> Result<()> {
        debug!(
            "updating result: id = {}, job_id = {}, agent_id = {}",
            result.id, result.job_id, result.agent_id
        );
        result
            .save_changes::<AssignedJob>(self.conn)
            .map_err(with_err_ctx(format!("Can't update result {result:?}")))?;
        Ok(())
    }
}

fn with_err_ctx(msg: impl AsRef<str>) -> impl Fn(DslError) -> Error {
    move |err| Error::DBErrorCtx(format!("{}, reason: {err}", msg.as_ref()))
}
