use diesel::result::Error as DslError;
use thiserror::Error;
use u_lib::{ufs, UError};
use warp::{
    http::StatusCode,
    reject::Reject,
    reply::{with_status, Response},
    Reply,
};

#[derive(Error, Debug)]
pub enum Error {
    #[error("Configs error: {0}")]
    ConfigError(#[from] u_lib::config::Error),

    #[error("Processing error: {0}")]
    ProcessingError(String),

    #[error(transparent)]
    DBError(#[from] DslError),

    #[error("DB error: {0}")]
    DBErrorCtx(String),

    #[error("Deadpool error: {0}")]
    DeadpoolError(#[from] deadpool_diesel::PoolError),

    #[error(transparent)]
    FSError(#[from] ufs::Error),

    #[error("Job cannot be ran on this platform. Expected: {0}, got: {1}")]
    InsuitablePlatform(String, String),

    #[error("{0}\nContext: {1}")]
    Contexted(Box<Error>, String),

    #[error(transparent)]
    UError(#[from] UError),

    #[error("Runtime error: {0}")]
    Runtime(String),
}

impl Reject for Error {}

pub struct RejResponse {
    message: String,
    status: StatusCode,
}

impl RejResponse {
    pub fn not_found(msg: impl Into<String>) -> Self {
        Self {
            message: msg.into(),
            status: StatusCode::NOT_FOUND,
        }
    }

    pub fn bad_request(msg: impl Into<String>) -> Self {
        Self {
            message: msg.into(),
            status: StatusCode::BAD_REQUEST,
        }
    }

    pub fn internal() -> Self {
        Self {
            message: "INTERNAL SERVER ERROR".to_string(),
            status: StatusCode::INTERNAL_SERVER_ERROR,
        }
    }
}

impl Reply for RejResponse {
    fn into_response(self) -> Response {
        with_status(self.message, self.status).into_response()
    }
}

impl From<anyhow::Error> for Error {
    fn from(e: anyhow::Error) -> Self {
        let ctx = e
            .chain()
            .rev()
            .skip(1)
            .map(|cause| format!("ctx: {}", cause))
            .collect::<Vec<_>>()
            .join("\n");

        match e.downcast::<Error>() {
            Ok(err) => Error::Contexted(Box::new(err), ctx),
            Err(err) => match err.downcast::<ufs::Error>() {
                Ok(err) => Error::Contexted(Box::new(Error::FSError(err)), ctx),
                Err(err) => Error::Runtime(err.to_string()),
            },
        }
    }
}
