use std::sync::Arc;

use crate::db::{PgRepo, UDB};
use crate::error::Error;
use serde::Deserialize;
use u_lib::{api::api_types, messaging::Reportable, models::*, types::Id};
use warp::reject::not_found;
use warp::Rejection;

type EndpResult<T> = Result<T, Rejection>;

#[derive(Deserialize)]
pub struct PayloadFlags {
    brief: Brief,
}

pub struct Endpoints;

impl Endpoints {
    pub async fn get_agents(repo: Arc<PgRepo>, id: Option<Id>) -> EndpResult<api_types::GetAgents> {
        repo.interact(move |mut db| {
            Ok(match id {
                Some(id) => {
                    if let Some(agent) = db.get_agent(id)? {
                        vec![agent]
                    } else {
                        vec![]
                    }
                }
                None => db.get_agents()?,
            })
        })
        .await
        .map_err(From::from)
    }

    pub async fn get_job(
        repo: Arc<PgRepo>,
        id: Id,
        params: Option<PayloadFlags>,
    ) -> EndpResult<api_types::GetJob> {
        let Some(mut job) = repo.interact(move |mut db| db.get_job(id)).await? else {
            return Err(not_found());
        };

        Ok(match params.map(|p| p.brief) {
            Some(Brief::Yes) => job,
            Some(Brief::Auto) | None => {
                if let Some(payload) = &mut job.payload {
                    payload.maybe_join_payload().map_err(Error::from)?;
                }
                job
            }
            Some(Brief::No) => {
                if let Some(payload) = &mut job.payload {
                    payload.join_payload().map_err(Error::from)?;
                }
                job
            }
        })
    }

    pub async fn get_jobs(repo: Arc<PgRepo>) -> EndpResult<api_types::GetJobs> {
        repo.interact(move |mut db| db.get_jobs())
            .await
            .map_err(From::from)
    }

    pub async fn get_assigned_jobs(
        repo: Arc<PgRepo>,
        id: Option<Id>,
    ) -> EndpResult<api_types::GetAgentJobs> {
        repo.interact(move |mut db| db.get_assigned_jobs(id, false))
            .await
            .map_err(From::from)
    }

    pub async fn get_payloads(repo: Arc<PgRepo>) -> EndpResult<api_types::GetPayloads> {
        repo.interact(move |mut db| db.get_payloads())
            .await
            .map_err(From::from)
    }

    pub async fn get_payload(
        repo: Arc<PgRepo>,
        name_or_id: String,
        params: Option<PayloadFlags>,
    ) -> EndpResult<api_types::GetPayload> {
        let mut payload = match repo
            .interact(move |mut db| match Id::parse_str(&name_or_id) {
                Ok(id) => db.get_payload(id),
                Err(_) => db.get_payload_by_name(name_or_id),
            })
            .await?
        {
            Some(p) => p,
            None => return Err(not_found()),
        };

        Ok(match params.map(|p| p.brief) {
            Some(Brief::Yes) => {
                payload.data = None;
                payload
            }
            None | Some(Brief::Auto) => {
                payload.maybe_join_payload().map_err(Error::from)?;
                payload
            }
            _ => {
                payload.join_payload().map_err(Error::from)?;
                payload
            }
        })
    }

    pub async fn get_personal_jobs(
        repo: Arc<PgRepo>,
        agent_id: Id,
    ) -> EndpResult<api_types::GetPersonalJobs> {
        repo.transaction(move |mut db| {
            let agent = db.get_agent(agent_id)?;
            match agent {
                Some(mut agent) => {
                    agent.touch();
                    db.upsert_agent(&agent)?;
                }
                None => {
                    let mut new_agent = Agent::empty();
                    new_agent.id = agent_id;

                    db.upsert_agent(&new_agent)?;

                    let job = db
                        .get_job_by_alias("agent_hello")?
                        .expect("agent_hello job not found");

                    let assigned_job = AssignedJobById {
                        agent_id,
                        job_id: job.meta.id,
                        ..Default::default()
                    };
                    db.assign_jobs(&[assigned_job])?;
                }
            }

            let assigned_jobs = db.get_assigned_jobs(Some(agent_id), true)?;

            for job in &assigned_jobs {
                db.update_job_status(job.id, JobState::Running)?;
            }

            Ok(assigned_jobs
                .into_iter()
                .map(|j| AssignedJobById::from(&j))
                .collect())
        })
        .await
        .map_err(From::from)
    }

    pub async fn upload_jobs(
        repo: Arc<PgRepo>,
        jobs: Vec<Job>,
    ) -> EndpResult<api_types::UploadJobs> {
        let mut checked_jobs = vec![];
        for mut job in jobs {
            if let Some(payload) = &mut job.payload {
                payload.maybe_split_payload().map_err(Error::from)?;
            } else if let Some(pld_id) = job.meta.payload_id {
                if !repo
                    .interact(move |mut db| db.payload_exists(pld_id))
                    .await?
                {
                    Err(Error::ProcessingError(format!(
                        "Payload {pld_id} not found"
                    )))?
                }
            }
            checked_jobs.push(job)
        }

        let (jobs, payloads_opt): (Vec<_>, Vec<_>) = checked_jobs
            .into_iter()
            .map(|j| (j.meta, j.payload))
            .unzip();

        let payloads = payloads_opt
            .into_iter()
            .filter_map(|p| p)
            .collect::<Vec<_>>();

        repo.transaction(move |mut db| {
            for payload in payloads {
                db.insert_payload(&payload)?;
            }
            db.insert_jobs(&jobs)
        })
        .await
        .map_err(From::from)
    }

    pub async fn upload_payload(
        repo: Arc<PgRepo>,
        raw_payload: RawPayload,
    ) -> EndpResult<api_types::UploadPayloads> {
        let payloads = raw_payload.into_payload().map_err(Error::from)?;

        repo.interact(move |mut db| db.insert_payload(&payloads))
            .await
            .map_err(From::from)
    }

    pub async fn del(repo: Arc<PgRepo>, id: Id) -> EndpResult<()> {
        repo.transaction(move |mut db| {
            [
                UDB::del_agents,
                UDB::del_jobs,
                UDB::del_results,
                UDB::del_payloads,
            ]
            .iter()
            .map(|f| f(&mut db, &[id]))
            .collect::<Result<(), Error>>()
        })
        .await
        .map_err(From::from)
    }

    pub async fn assign_jobs(
        repo: Arc<PgRepo>,
        assigned_jobs: Vec<AssignedJobById>,
    ) -> EndpResult<()> {
        repo.transaction(move |mut db| {
            db.assign_jobs(&assigned_jobs)?;
            Ok(())
        })
        .await
        .map_err(From::from)
    }

    pub async fn report(
        repo: Arc<PgRepo>,
        msg: Vec<Reportable>,
        agent_id: Id,
    ) -> EndpResult<api_types::Report> {
        repo.transaction(move |mut db| {
            for entry in msg {
                match entry {
                    Reportable::Assigned(mut result) => {
                        let result_agent_id = &result.agent_id;
                        if agent_id != *result_agent_id {
                            warn!(
                                "Agent ids are not equal! actual id: {agent_id}, \
                                id from job: {result_agent_id}"
                            );
                            continue;
                        }
                        result.touch();

                        info!("agent {agent_id} updated job {}", result.id);

                        match result.exec_type {
                            JobType::Init => {
                                result.state = JobState::Finished;

                                let mut agent: Agent = match result.deserialize() {
                                    Ok(a) => a,
                                    Err(e) => {
                                        error!(
                                            "Error deserializing agent \
                                                    data from {agent_id}: {e}"
                                        );
                                        continue;
                                    }
                                };
                                agent.state = AgentState::Active;
                                db.upsert_agent(&agent)?;
                            }
                            JobType::Shell => result.state = JobState::Finished,
                            JobType::Stats => result.state = JobState::Finished,
                            JobType::Terminate => (),
                            JobType::Update => (),
                        }
                        db.update_result(&result)?;
                    }
                    Reportable::Error(e) => {
                        error!("agent {agent_id} reported: {e}");
                    }
                    Reportable::Dummy => (),
                }
            }
            Ok(())
        })
        .await
        .map_err(From::from)
    }

    pub async fn update_agent(
        repo: Arc<PgRepo>,
        agent: Agent,
    ) -> EndpResult<api_types::UpdateAgent> {
        repo.interact(move |mut db| db.upsert_agent(&agent))
            .await
            .map_err(From::from)
    }

    pub async fn update_job(repo: Arc<PgRepo>, job: JobMeta) -> EndpResult<api_types::UpdateJob> {
        repo.interact(move |mut db| db.update_job(&job.validate()?))
            .await
            .map_err(From::from)
    }

    pub async fn update_assigned_job(
        repo: Arc<PgRepo>,
        assigned: AssignedJob,
    ) -> EndpResult<api_types::UpdateResult> {
        repo.interact(move |mut db| db.update_result(&assigned))
            .await
            .map_err(From::from)
    }

    pub async fn update_payload(
        repo: Arc<PgRepo>,
        payload: Payload,
    ) -> EndpResult<api_types::UpdatePayload> {
        debug!("update payload: {payload:?}");
        match payload.data {
            Some(data) => {
                let mut well_formed_payload =
                    Payload::from_data(data, Some(payload.name)).map_err(Error::from)?;
                well_formed_payload.id = payload.id;

                repo.interact(move |mut db| db.update_payload(&well_formed_payload))
                    .await
                    .map_err(From::from)
            }
            None => repo
                .interact(move |mut db| db.update_payload(&payload))
                .await
                .map_err(From::from),
        }
    }
}
