use crate::{
    cache::JobCache,
    combined_result::CombinedResult,
    config::get_self_id,
    models::{Agent, AssignedJob, AssignedJobById, Job, JobType, PreparedJob, RawJob, Stats},
    proc_output::ProcOutput,
    scheduler::SCHEDULER,
    u_runner::{ExecResult, IdentifiableFuture, URunner},
    ufs,
};
use std::collections::HashMap;
use std::process::exit;
use tokio::process::Command;

pub struct AnonymousJobBatch {
    runner: URunner,
    is_running: bool,
}

impl AnonymousJobBatch {
    pub fn from_prepared_jobs(jobs: impl IntoIterator<Item = PreparedJob>) -> Self {
        let mut runner = URunner::new();
        for job in jobs {
            runner.push(IdentifiableFuture::from_fut_with_ident(
                job.ids.id,
                run_assigned_job(job),
            ));
        }
        Self {
            runner,
            is_running: false,
        }
    }

    pub fn from_jobs(jobs: impl IntoIterator<Item = Job>) -> Self {
        let jobs_ids: Vec<_> = jobs
            .into_iter()
            .map(|job| PreparedJob {
                ids: AssignedJobById {
                    job_id: job.meta.id,
                    ..Default::default()
                },
                job,
            })
            .collect();
        AnonymousJobBatch::from_prepared_jobs(jobs_ids)
    }

    /// Spawn jobs
    pub async fn spawn(mut self) -> Self {
        debug!("spawning jobs");
        self.runner = self.runner.spawn().await;
        self.is_running = true;
        self
    }

    /// Spawn jobs and wait for result
    pub async fn wait(self) -> Vec<ExecResult> {
        let runner = if !self.is_running {
            self.spawn().await.runner
        } else {
            self.runner
        };
        runner.wait().await
    }

    /// Spawn one job and wait for result
    pub async fn wait_one(self) -> ExecResult {
        self.wait().await.pop().unwrap()
    }
}

/// Store jobs and get results by name
pub struct NamedJobBatch<const FINISHED: bool = false> {
    runner: Option<AnonymousJobBatch>,
    job_names: Vec<String>,
    results: HashMap<String, ExecResult>,
}

impl NamedJobBatch {
    pub fn from_shell(named_jobs: Vec<(&'static str, &'static str)>) -> CombinedResult<Self> {
        let mut result = CombinedResult::new();
        let jobs: Vec<_> = named_jobs
            .into_iter()
            .filter_map(|(alias, cmd)| {
                match RawJob::default()
                    .with_shell(cmd)
                    .with_alias(alias)
                    .try_into_job()
                {
                    Ok(jpm) => Some(jpm),
                    Err(e) => {
                        result.push_err(e);
                        None
                    }
                }
            })
            .collect();
        result.push_ok(Self::from_meta(jobs));
        result
    }

    pub fn from_meta(named_jobs: Vec<Job>) -> Self {
        let (job_names, jobs): (Vec<_>, Vec<_>) = named_jobs
            .into_iter()
            .map(|job| (job.meta.alias.clone().unwrap(), job))
            .unzip();
        Self {
            runner: Some(AnonymousJobBatch::from_jobs(jobs)),
            job_names,
            results: HashMap::new(),
        }
    }

    pub async fn wait(mut self) -> NamedJobBatch<true> {
        let results = self.runner.take().unwrap().wait().await;
        for (name, result) in self.job_names.into_iter().zip(results.into_iter()) {
            self.results.insert(name, result);
        }

        NamedJobBatch::<true> {
            runner: None,
            job_names: vec![],
            results: self.results,
        }
    }
}

impl NamedJobBatch<true> {
    pub fn pop_opt(&mut self, name: &'static str) -> Option<ExecResult> {
        self.results.remove(name)
    }

    pub fn pop(&mut self, name: &'static str) -> ExecResult {
        self.pop_opt(name).unwrap()
    }
}

pub async fn run_assigned_job(prepared_job: PreparedJob) -> ExecResult {
    let mut result = AssignedJob::from(&prepared_job);
    let Job { meta, payload } = prepared_job.job;
    match meta.exec_type {
        JobType::Shell => {
            let (argv, _prepared_payload) = {
                if let Some(payload) = payload {
                    let (prep_exec, prep_exec_path) = payload.prepare_executable()?;
                    let argv_with_exec = meta.argv.replace("{}", &prep_exec_path);
                    (argv_with_exec, Some(prep_exec))
                } else {
                    (meta.argv, None)
                }
            };

            let mut split_cmd = shlex::split(&argv).unwrap().into_iter();
            let cmd = split_cmd.nth(0).unwrap();
            let args = split_cmd.collect::<Vec<String>>();
            let cmd_result = Command::new(cmd)
                .kill_on_drop(true)
                .args(args)
                .output()
                .await;
            let (data, retcode) = match cmd_result {
                Ok(output) => (
                    ProcOutput::from_output(&output).into_vec(),
                    output.status.code(),
                ),
                Err(e) => (
                    ProcOutput::new()
                        .set_stderr(e.to_string().into_bytes())
                        .into_vec(),
                    None,
                ),
            };
            result.result = Some(data);
            result.retcode = retcode;
        }
        JobType::Init => {
            let agent = Agent::gather().await;
            debug!("gathered info from agent: {agent:?}");
            result.set_result(&agent);
            result.retcode = Some(0);
        }
        JobType::Stats => {
            let stats = Stats {
                agent_id: get_self_id(),
                running_jobs: URunner::stats().await,
                scheduled_jobs: SCHEDULER.stats().await,
                cached_jobs: JobCache::stats(),
                cached_payloads: ufs::stats(),
            };
            result.set_result(&stats);
            result.retcode = Some(0)
        }
        JobType::Update => (),
        JobType::Terminate => exit(0),
    };
    Ok(result)
}

#[cfg(test)]
mod tests {
    use crate::{
        jobs::{AnonymousJobBatch, NamedJobBatch},
        models::{misc::JobType, RawJob},
        unwrap_enum, UError,
    };
    use std::time::SystemTime;

    type TestResult<R = ()> = Result<R, Box<dyn std::error::Error>>;

    #[tokio::test]
    async fn test_is_really_async() {
        const SLEEP_SECS: u64 = 1;
        let job = RawJob::from_shell(format!("sleep {SLEEP_SECS}")).unwrap();
        let sleep_jobs = (0..50).map(|_| job.clone()).collect::<Vec<_>>();

        let now = SystemTime::now();
        AnonymousJobBatch::from_jobs(sleep_jobs).wait().await;
        assert!(now.elapsed().unwrap().as_secs() < SLEEP_SECS + 2)
    }

    #[rstest]
    #[case::sh_payload(
        "/bin/sh {}",
        Some(b"echo test01 > /tmp/asd; cat /tmp/asd".as_slice()),
        "test01"
    )]
    #[case::python_cmd(r#"/usr/bin/python3 -c 'print("test02")'"#, None, "test02")]
    #[case::sh_multiline_payload(
        "/{}",
        Some(
    br#"#!/bin/sh
    TMPPATH=/tmp/lol
    mkdir -p $TMPPATH
    echo test03 > $TMPPATH/t
    cat $TMPPATH/t
    rm -rf $TMPPATH"#.as_slice()
        ),
        "test03"
    )]
    #[case::standalone_binary_with_args(
        "/{} 'some msg as arg'",
        Some(include_bytes!("../tests/fixtures/echoer").as_slice()),
        "some msg as arg"
    )]
    #[tokio::test]
    async fn test_shell_job(
        #[case] cmd: &str,
        #[case] payload: Option<&[u8]>,
        #[case] expected_result: &str,
    ) -> TestResult {
        let mut raw_job = RawJob::default().with_shell(cmd);
        if let Some(p) = payload {
            raw_job = raw_job.with_raw_payload(p);
        }
        let job = raw_job.try_into_job().unwrap();
        let result = AnonymousJobBatch::from_jobs([job])
            .wait_one()
            .await
            .unwrap();
        let result = result.to_str_result();
        assert_eq!(result.trim(), expected_result);
        Ok(())
    }

    #[tokio::test]
    async fn test_complex_load() -> TestResult {
        const SLEEP_SECS: u64 = 1;
        let now = SystemTime::now();
        let longest_job = RawJob::from_shell(format!("sleep {}", SLEEP_SECS)).unwrap();
        let longest_job = AnonymousJobBatch::from_jobs([longest_job]).spawn().await;
        let ls = AnonymousJobBatch::from_jobs([RawJob::from_shell("ls").unwrap()])
            .wait_one()
            .await
            .unwrap();

        assert_eq!(ls.retcode.unwrap(), 0);

        let folders = ls.to_str_result();
        let subfolders_jobs = folders
            .lines()
            .map(|f| RawJob::from_shell(format!("ls {f}")).unwrap())
            .collect::<Vec<_>>();

        let ls_subfolders = AnonymousJobBatch::from_jobs(subfolders_jobs).wait().await;

        for result in ls_subfolders {
            assert_eq!(result.unwrap().retcode.unwrap(), 0);
        }

        longest_job.wait().await;
        assert_eq!(now.elapsed().unwrap().as_secs(), SLEEP_SECS);
        Ok(())
    }
    /*
    #[tokio::test]
    async fn test_exec_multiple_jobs_nowait() -> UResult<()> {
        const REPEATS: usize = 10;
        let job = JobMeta::from_shell("whoami");
        let sleep_jobs: Vec<JobMeta> = (0..=REPEATS).map(|_| job.clone()).collect();
        build_jobs(sleep_jobs).spawn().await;
        let mut completed = 0;
        while completed < REPEATS {
            let c = pop_completed().await.len();
            if c > 0 {
                completed += c;
                println!("{}", c);
            }
        }
        Ok(())
    }
    */
    #[tokio::test]
    async fn test_failing_shell_job() -> TestResult {
        let job = RawJob::from_shell("lol_kek_puk").unwrap();
        let job_result = AnonymousJobBatch::from_jobs([job])
            .wait_one()
            .await
            .unwrap();
        let output = job_result.to_str_result();
        assert!(output.contains("No such file"));
        assert!(job_result.retcode.is_none());
        Ok(())
    }

    #[rstest]
    #[case::no_binary("/bin/bash {}", None, "contains executable")]
    #[case::no_path_to_binary("/bin/bash", Some(b"whoami".as_slice()), "contains no executable")]
    #[tokio::test]
    async fn test_job_building_failed(
        #[case] cmd: &str,
        #[case] payload: Option<&[u8]>,
        #[case] err_str: &str,
    ) -> TestResult {
        let mut raw_job = RawJob::default().with_shell(cmd);
        if let Some(p) = payload {
            raw_job = raw_job.with_raw_payload(p);
        }
        let err = raw_job.try_into_job().unwrap_err();
        let err_msg = unwrap_enum!(err, UError::JobBuildError);
        assert!(err_msg.contains(err_str));
        Ok(())
    }

    #[tokio::test]
    async fn test_different_job_types() -> TestResult {
        let mut jobs = NamedJobBatch::from_meta(vec![
            RawJob::default()
                .with_shell("sleep 3")
                .with_alias("sleeper")
                .try_into_job()
                .unwrap(),
            RawJob::default()
                .with_type(JobType::Init)
                .with_alias("gatherer")
                .try_into_job()
                .unwrap(),
        ])
        .wait()
        .await;
        let gathered = jobs.pop("gatherer").unwrap();
        assert_eq!(gathered.alias.unwrap(), "gatherer");
        Ok(())
    }
}
