use std::fmt;

use super::JobType;
#[cfg(feature = "server")]
use crate::models::schema::*;
use crate::models::Payload;
use crate::platform;
use crate::types::Id;
use crate::{UError, UResult};
use cron::Schedule;
#[cfg(feature = "server")]
use diesel::{Identifiable, Insertable, Queryable};
use serde::{Deserialize, Serialize};
use std::borrow::Cow;

#[derive(Clone, Serialize, Deserialize, PartialEq, Eq)]
#[cfg_attr(
    feature = "server",
    derive(Queryable, Identifiable, Insertable, AsChangeset),
    diesel(table_name = jobs),
    diesel(treat_none_as_null = true)
)]
pub struct JobMeta {
    #[serde(default)]
    pub alias: Option<String>,

    /// string like `bash -c {} -a 1 --arg2`,
    /// where {} is replaced by executable's tmp path
    #[serde(default)]
    pub argv: String,

    #[serde(default = "Id::new_v4")]
    pub id: Id,

    #[serde(default)]
    pub exec_type: JobType,

    /// target triple
    #[serde(default = "platform::current_as_string")]
    pub target_platforms: String,

    #[serde(default)]
    pub payload_id: Option<Id>,

    /// cron-like string
    #[serde(default)]
    pub schedule: Option<String>,
}

impl Default for JobMeta {
    fn default() -> Self {
        Self {
            alias: None,
            argv: String::new(),
            id: Id::new_v4(),
            exec_type: JobType::default(),
            target_platforms: String::new(),
            payload_id: None,
            schedule: None,
        }
    }
}

impl fmt::Debug for JobMeta {
    fn fmt(&self, f: &mut fmt::Formatter<'_>) -> fmt::Result {
        f.debug_struct("JobMeta")
            .field("alias", &self.alias)
            .field("argv", &self.argv)
            .field("id", &self.id.to_string())
            .field("exec_type", &self.exec_type)
            .field("target_platforms", &self.target_platforms)
            .field("payload_id", &self.payload_id.map(|id| id.to_string()))
            .field("schedule", &self.schedule)
            .finish()
    }
}

impl JobMeta {
    pub fn validate(mut self) -> UResult<Self> {
        fn mk_err(msg: impl Into<String>) -> UError {
            UError::JobBuildError(msg.into())
        }

        const ARGV_STR_LEN: usize = 2048;

        if self.argv.is_empty() {
            // TODO: fix detecting
            self.argv = String::from("echo 'hello, world!'")
        } else if self.argv.len() > ARGV_STR_LEN {
            return Err(mk_err(format!(
                "argv length limit ({ARGV_STR_LEN}) exceeded"
            )));
        }

        let argv_parts = shlex::split(&self.argv).ok_or_else(|| mk_err("Shlex failed"))?;
        let empty_err = mk_err("Empty argv");

        if argv_parts.get(0).ok_or(empty_err.clone())?.is_empty() {
            return Err(empty_err);
        }

        if self.payload_id.is_some() && !self.argv.contains("{}") {
            return Err(mk_err("Argv contains no executable placeholder"));
        }

        if self.payload_id.is_none() && self.argv.contains("{}") {
            return Err(mk_err(
                "No payload provided, but argv contains executable placeholder",
            ));
        }

        if self.target_platforms.is_empty() {
            self.target_platforms = "*".to_string();
        }

        if !platform::is_valid_glob(&self.target_platforms) {
            return Err(mk_err(format!(
                "Unknown platform '{}'",
                self.target_platforms
            )));
        }

        if let Some(schedule) = &self.schedule {
            Schedule::try_from(schedule.as_str())
                .map_err(|e| mk_err(format!("bad schedule: {e}")))?;
        }

        Ok(self)
    }
}

/// An abstract valid-constructed job that can be assigned to any agent.
/// Contains job meta and payload meta
#[derive(Debug, Clone, Serialize, Deserialize, PartialEq, Eq)]
pub struct Job {
    pub meta: JobMeta,
    pub payload: Option<Payload>,
}

/// Job that has only been deserialized without checks
#[derive(Serialize, Deserialize, Clone, Default)]
pub struct RawJob<'p> {
    #[serde(default)]
    payload_path: Option<String>,

    #[serde(default)]
    raw_payload: Option<Cow<'p, [u8]>>,

    #[serde(default, flatten)]
    meta: JobMeta,
}

impl From<Job> for RawJob<'_> {
    fn from(job: Job) -> Self {
        let Job { meta, payload } = job;

        RawJob {
            payload_path: payload.map(|m| m.name),
            raw_payload: None,
            meta,
        }
    }
}

impl<'p> RawJob<'p> {
    pub fn try_into_job(self) -> UResult<Job> {
        Job::try_from(self)
    }

    pub fn from_shell(cmd: impl Into<String>) -> UResult<Job> {
        Self::default().with_shell(cmd).try_into_job()
    }

    pub fn with_shell(mut self, shell_cmd: impl Into<String>) -> Self {
        self.meta.argv = shell_cmd.into();
        self.meta.exec_type = JobType::Shell;
        self
    }

    pub fn with_alias(mut self, alias: impl Into<String>) -> Self {
        self.meta.alias = Some(alias.into());
        self
    }

    pub fn with_type(mut self, e_type: JobType) -> Self {
        self.meta.exec_type = e_type;
        self
    }

    pub fn with_target_platforms(mut self, platform: impl Into<String>) -> Self {
        self.meta.target_platforms = platform.into();
        self
    }

    pub fn with_raw_payload(mut self, raw_payload: impl AsPayload<'p>) -> Self {
        self.raw_payload = Some(raw_payload.as_payload());
        self.payload_path = None;
        self
    }

    pub fn with_payload_path(mut self, path: impl Into<String>) -> Self {
        self.payload_path = Some(path.into());
        self.raw_payload = None;
        self
    }
}

impl TryFrom<RawJob<'_>> for Job {
    type Error = UError;

    fn try_from(mut raw: RawJob) -> Result<Self, Self::Error> {
        if raw.raw_payload.is_some() && raw.payload_path.is_some() {
            return Err(UError::JobBuildError(
                "Can't use both raw payload with payload path".to_string(),
            ));
        }

        let payload = {
            let payload_from_path = raw
                .payload_path
                .as_ref()
                .map(|path| Payload::from_path(path))
                .transpose()?;

            if payload_from_path.is_none() {
                raw.raw_payload
                    .as_ref()
                    .map(|data| Payload::from_data(data, None))
                    .transpose()?
            } else {
                payload_from_path
            }
        };

        raw.meta.payload_id = payload.as_ref().map(|p| p.id).or(raw.meta.payload_id);

        Ok(Job {
            meta: raw.meta.validate()?,
            payload,
        })
    }
}

pub trait AsPayload<'p> {
    fn as_payload(&self) -> Cow<'p, [u8]>;
}

impl<'p> AsPayload<'p> for &'p str {
    fn as_payload(&self) -> Cow<'p, [u8]> {
        Cow::Borrowed(self.as_bytes())
    }
}

impl<'p, const N: usize> AsPayload<'p> for &'p [u8; N] {
    fn as_payload(&self) -> Cow<'p, [u8]> {
        Cow::Borrowed(*self)
    }
}

impl<'p> AsPayload<'p> for &'p [u8] {
    fn as_payload(&self) -> Cow<'p, [u8]> {
        Cow::Borrowed(self)
    }
}
