use std::process::Output;

#[derive(Clone, Debug)]
pub struct ProcOutput {
    pub stdout: Vec<u8>,
    pub stderr: Vec<u8>,
}

impl ProcOutput {
    const STDERR_DELIMETER: &'static [u8] = b"[STDERR]\n";

    pub fn from_output(output: &Output) -> Self {
        Self::new()
            .set_stdout(output.stdout.to_vec())
            .set_stderr(output.stderr.to_vec())
    }

    pub fn new() -> Self {
        Self {
            stdout: vec![],
            stderr: vec![],
        }
    }

    pub fn set_stdout(mut self, data: Vec<u8>) -> Self {
        self.stdout = data;
        self
    }

    pub fn set_stderr(mut self, data: Vec<u8>) -> Self {
        self.stderr = data;
        self
    }

    pub fn get_stdout(&self) -> &[u8] {
        &self.stdout
    }

    pub fn get_stderr(&self) -> &[u8] {
        &self.stderr
    }

    pub fn into_vec(self) -> Vec<u8> {
        let mut result: Vec<u8> = vec![];
        let stdout_is_empty = self.stdout.is_empty();
        if !stdout_is_empty {
            result.extend(self.stdout);
        }

        if !self.stderr.is_empty() {
            if !stdout_is_empty && *result.last().unwrap() != b'\n' {
                result.extend(b"\n");
            }
            result.extend(Self::STDERR_DELIMETER);
            result.extend(self.stderr);
        }
        result
    }

    pub fn into_readable(self) -> String {
        String::from_utf8_lossy(&self.into_vec()).to_string()
    }

    pub fn from_raw_proc_output(raw: &[u8]) -> Self {
        let stderr_delim_len = Self::STDERR_DELIMETER.len();
        let split_pos = raw
            .windows(stderr_delim_len)
            .position(|w| w == Self::STDERR_DELIMETER)
            .unwrap_or(raw.len());

        let (stdout, stderr) = raw.split_at(split_pos);
        let result = Self::new().set_stdout(stdout.to_vec());
        if stderr.len() >= stderr_delim_len {
            result.set_stderr(stderr[stderr_delim_len..].to_vec())
        } else {
            result
        }
    }
}

#[cfg(test)]
mod tests {
    use crate::{conv::bytes_to_string, proc_output::ProcOutput};
    use std::str;

    const STDERR_DELIMETER: &'static str =
        unsafe { str::from_utf8_unchecked(ProcOutput::STDERR_DELIMETER) };

    #[rstest]
    #[case::stdout_stderr(
        "lol",
        "kek",
        &format!("lol\n{}kek", STDERR_DELIMETER)
    )]
    #[case::stderr(
        "",
        "kek",
        &format!("{}kek", STDERR_DELIMETER)
    )]
    fn test_to_combined(#[case] stdout: &str, #[case] stderr: &str, #[case] expected_result: &str) {
        let output = ProcOutput::new()
            .set_stdout(stdout.as_bytes().to_vec())
            .set_stderr(stderr.as_bytes().to_vec());
        assert_eq!(bytes_to_string(&output.into_vec()), expected_result)
    }

    #[rstest]
    #[case::stdout_stderr(
        &format!("lal\n{}kik", STDERR_DELIMETER),
    )]
    #[case::stdout(
        &format!("qeq"),
    )]
    #[case::stderr(
        &format!("{}vev", STDERR_DELIMETER),
    )]
    fn test_from_combined(#[case] expected_result: &str) {
        let output = ProcOutput::from_raw_proc_output(expected_result.as_bytes());
        assert_eq!(bytes_to_string(&output.into_vec()).trim(), expected_result);
    }
}
