mod entry;

use std::sync::Arc;
use std::time::Duration;

use chrono::Utc;
use cron::Schedule;
use entry::Entry;
use once_cell::sync::Lazy;
use tokio::sync::Mutex;
use uuid::Uuid;

use crate::jobs::AnonymousJobBatch;
use crate::types::Id;

use self::entry::EntryType;

pub use entry::EntryStat;

pub static SCHEDULER: Lazy<AsyncScheduler> = Lazy::new(|| AsyncScheduler::new());

#[derive(Clone)]
pub struct AsyncScheduler {
    entries: Arc<Mutex<Vec<Entry>>>,
}

impl AsyncScheduler {
    pub fn new() -> AsyncScheduler {
        AsyncScheduler {
            entries: Arc::new(Mutex::new(Vec::new())),
        }
    }

    pub async fn start_blocking(&self) -> ! {
        for entry in self.entries.lock().await.iter_mut() {
            entry.set_next_run_time();
        }

        loop {
            let mut entries = self.entries.lock().await;
            entries.sort_by(|b, a| b.next.cmp(&a.next));

            let wait_duration = if let Some(entry) = entries.first() {
                let wait_millis = (entry.next.as_ref().unwrap().timestamp_millis() as u64)
                    .saturating_sub(Utc::now().timestamp_millis() as u64);

                Duration::from_millis(wait_millis)
            } else {
                Duration::from_secs(1)
            };

            drop(entries);
            tokio::time::sleep(wait_duration).await;

            let mut entries = self.entries.lock().await;
            let mut job_batch = vec![];

            for entry in &mut *entries {
                match entry.next.as_ref() {
                    Some(next) => {
                        if next.gt(&Utc::now()) {
                            break;
                        }
                        match &entry.runnable {
                            EntryType::Common(runnable) => {
                                let fut = runnable.call();
                                tokio::spawn(fut);
                            }
                            EntryType::URunner(runnable) => {
                                debug!("starting assigned job {:?}", runnable.ids);
                                job_batch.push(runnable.clone())
                            }
                        }
                        entry.set_next_run_time();
                    }
                    None => {}
                }
            }

            AnonymousJobBatch::from_prepared_jobs(job_batch)
                .spawn()
                .await;

            entries.retain(|e| e.schedule.is_some());
        }
    }

    pub async fn add_job(&self, schedule: Option<Schedule>, runnable: impl Into<EntryType>) -> Id {
        let entry_id = Uuid::new_v4();
        let runnable = runnable.into();
        let mut entry = Entry {
            entry_id,
            schedule,
            next: None,
            runnable,
        };

        entry.set_next_run_time();
        self.entries.lock().await.push(entry);

        entry_id
    }

    pub async fn del_job(&self, entry_id: Id) {
        self.entries.lock().await.retain(|e| e.entry_id != entry_id);
    }

    pub async fn stats(&self) -> Vec<EntryStat> {
        self.entries
            .lock()
            .await
            .iter()
            .map(|entry| entry.as_stat())
            .collect()
    }

    pub async fn start(&self) {
        let cloned = self.clone();
        tokio::spawn(async move {
            cloned.start_blocking().await;
        });
    }
}

#[cfg(test)]
mod tests {
    use std::time::Duration;
    use tokio::time::sleep;

    use crate::u_runner::IdentifiableFuture;

    use super::*;

    #[tokio::test(flavor = "multi_thread")]
    async fn scheduling() {
        let v = Arc::new(Mutex::new(0));
        let scheduler = AsyncScheduler::new();

        {
            let v = v.clone();
            scheduler
                .add_job(Some("*/1 * * * * * *".try_into().unwrap()), move || {
                    let v = v.clone();
                    IdentifiableFuture::from_fut_with_ident("testf", async move {
                        *v.lock().await += 1;
                    })
                })
                .await;
        }

        scheduler
            .add_job(Some("*/1 * * * * * *".try_into().unwrap()), move || {
                IdentifiableFuture::from_fut_with_ident("testf", async move {
                    println!("hello1");
                })
            })
            .await;

        scheduler
            .add_job(Some("*/1 * * * * * *".try_into().unwrap()), move || {
                IdentifiableFuture::from_fut_with_ident("testf", async move {
                    println!("hello2");
                })
            })
            .await;

        scheduler.start().await;

        sleep(Duration::from_secs_f32(0.1)).await; // wait for scheduler
        sleep(Duration::from_secs(1)).await;

        assert_eq!(*v.lock().await, 1);

        sleep(Duration::from_secs(1)).await;

        assert_eq!(*v.lock().await, 2);
    }
}
