use super::Error;
use once_cell::sync::Lazy;
use parking_lot::Mutex;
use serde::{Deserialize, Serialize};
use std::collections::HashMap;
use std::env::temp_dir;
use std::ffi::OsString;
use std::fs;
use std::path::PathBuf;

// index format: given_name -> payload_meta
type Index = HashMap<String, IndexFileMeta>;

static IDX_FILE_NAME: Lazy<PathBuf> = Lazy::new(|| temp_dir().join(".i"));
static INDEX: Lazy<Mutex<Index>> = Lazy::new(|| {
    let idx_name = &*IDX_FILE_NAME;
    let deserialized_idx = fs::read(idx_name)
        .map_err(|e| Error::new(e, idx_name))
        .and_then(|raw_idx| {
            bincode::deserialize::<Index>(&raw_idx).map_err(|e| Error::new(e, idx_name))
        });

    let idx = match deserialized_idx {
        Ok(idx) => idx,
        Err(e) => {
            error!("index loading failed: {e}");
            HashMap::new()
        }
    };
    Mutex::new(idx)
});

#[derive(Debug, Clone, Deserialize, Serialize)]
pub struct IndexFileMeta {
    extension: Option<OsString>,
    pub external: bool, // if file is present before adding to index
    pub hashsum: String,
    pub path: PathBuf, // actual file path
    pub size: u64,
}

impl IndexFileMeta {
    pub fn new(
        full_path: impl Into<PathBuf>,
        hashsum: String,
        external: bool,
    ) -> Result<Self, Error> {
        let full_path: PathBuf = full_path.into();
        let extension = full_path.extension().map(ToOwned::to_owned);
        let size = fs::metadata(&full_path)?.len();

        Ok(IndexFileMeta {
            path: full_path,
            extension,
            external,
            hashsum,
            size,
        })
    }
}

mod sync {
    use super::{Index, IDX_FILE_NAME};
    use std::fs;

    pub(super) fn deleted(index: &mut Index) {
        let files_to_delete: Vec<String> = index
            .iter()
            .filter_map(|(name, meta)| {
                if meta.path.exists() {
                    None
                } else {
                    Some(name.to_string())
                }
            })
            .collect();

        files_to_delete.into_iter().for_each(|f| {
            index.remove(&f);
        });
    }

    pub(super) fn index2fs(index: &Index) {
        let serialized_idx = bincode::serialize(index).expect("broken index");
        if let Err(e) = fs::write(&*IDX_FILE_NAME, serialized_idx) {
            error!("index dumping failed: {e}");
        }
    }
}

pub fn get(name: impl AsRef<str>) -> Option<IndexFileMeta> {
    let mut index = INDEX.lock();

    sync::deleted(&mut index);

    index.get(name.as_ref()).cloned()
}

pub fn get_by_hash(hash: impl AsRef<str>) -> Option<(String, IndexFileMeta)> {
    let mut index = INDEX.lock();

    sync::deleted(&mut index);

    index
        .iter()
        .find(|(_name, meta)| meta.hashsum == hash.as_ref())
        .map(|(n, m)| (n.to_owned(), m.clone()))
}

pub fn insert(name: impl Into<String>, meta: IndexFileMeta) {
    let mut index = INDEX.lock();

    sync::deleted(&mut index);

    index.insert(name.into(), meta);

    #[cfg(any(feature = "panel", feature = "server"))]
    sync::index2fs(&mut index);
}

pub fn remove(name: impl AsRef<str>) -> Option<IndexFileMeta> {
    let mut index = INDEX.lock();

    sync::deleted(&mut index);

    let result = index.remove(name.as_ref());

    #[cfg(any(feature = "panel", feature = "server"))]
    sync::index2fs(&mut index);

    result
}

pub fn stats() -> Vec<IndexFileMeta> {
    INDEX.lock().values().cloned().collect()
}
