/// This module is aiming to store (obfuscated?) payloads, get them by name,
/// rename, update, delete or prepare to execute via memfd_create (unix)
mod error;
mod index;

use anyhow::{Context, Result};
use std::env::temp_dir;
use std::fs;
use std::path::{Path, PathBuf};
use uuid::Uuid;

pub use error::Error;
pub use index::IndexFileMeta;

const OBFUSCATE: bool = cfg!(feature = "agent");

/// Check if file exists in index.
/// File may present in fs but not in index, fn will return false then.
pub fn exists_in_index(name: impl AsRef<str>) -> bool {
    read_meta(name).is_ok()
}

#[inline]
pub fn read_meta(name: impl AsRef<str>) -> Result<IndexFileMeta> {
    index::get(&name)
        .ok_or_else(|| Error::not_found(name.as_ref()))
        .context("meta")
}

/// Read file by index name
pub fn read(name: impl AsRef<str>) -> Result<Vec<u8>> {
    let meta = read_meta(&name).context("read_meta")?;
    fs::read(&meta.path)
        .map_err(|e| Error::new(e, name.as_ref()))
        .context("read")
}

/// Create file with generated name
pub fn create_anonymous(data: impl AsRef<[u8]>) -> Result<String> {
    if let Some((name, _)) = index::get_by_hash(hash_data(&data)) {
        return Ok(name);
    }

    let name = Uuid::new_v4().simple().to_string();

    put(&name, data).context("cr_anon")?;

    Ok(name)
}

/// Create new file and add to index.
///
/// If index already contains a file with the same contents, it doesn't duplicate it,
/// but adds an symlink.
pub fn put(name: impl AsRef<str>, data: impl AsRef<[u8]>) -> Result<()> {
    let name = name.as_ref();
    let data_hash = hash_data(&data);

    debug!("put: {name}");

    let path = match index::get_by_hash(&data_hash) {
        Some((_, meta)) => meta.path,
        None => {
            let path = {
                let exec_name = if OBFUSCATE {
                    PathBuf::from(Uuid::new_v4().simple().to_string())
                } else {
                    PathBuf::from(name)
                };

                let mut path = temp_dir();
                path.push(exec_name);
                path
            };

            debug!("put: path: {path:?}");

            fs::write(&path, &data)
                .map_err(|e| Error::new(e, name))
                .context("put_write")?;
            path
        }
    };
    index::insert(
        name,
        IndexFileMeta::new(path, data_hash, false).context("put_insert")?,
    );

    Ok(())
}

pub fn edit(name: impl AsRef<str>, data: impl AsRef<[u8]>) -> Result<()> {
    let meta = read_meta(&name).context("edit_meta")?;
    let data_hash = hash_data(&data);

    if meta.hashsum == data_hash {
        return Ok(());
    }

    fs::write(&meta.path, &data)
        .map_err(|e| Error::new(e, &meta.path))
        .context("edit_write")?;

    let new_meta = IndexFileMeta::new(meta.path, data_hash, meta.external).context("edit_nmeta")?;

    index::remove(&name);
    index::insert(name.as_ref(), new_meta);

    Ok(())
}

pub fn remove(name: impl AsRef<str>) -> Result<()> {
    let name = name.as_ref();
    match index::remove(name) {
        Some(value) if !value.external => fs::remove_file(value.path)
            .map_err(|e| Error::new(e, name))
            .context("remove"),
        _ => Ok(()),
    }
}

// todo: don't rename external files
pub fn rename(old_name: impl AsRef<str>, new_name: impl AsRef<str>) -> Result<()> {
    let old_name = old_name.as_ref();
    let new_name = new_name.as_ref();

    if old_name == new_name {
        return Ok(());
    }

    if !exists_in_index(old_name) {
        return Err(Error::not_found(old_name)).context("rename");
    }

    if exists_in_index(new_name) {
        return Err(Error::already_exists(new_name)).context("rename");
    }

    let mut value = index::remove(old_name).unwrap();

    if !OBFUSCATE {
        let old_path = value.path.clone();

        value.path.pop();
        value.path.push(new_name);

        fs::rename(old_path, &value.path)
            .map_err(|e| Error::new(e, &value.path))
            .context("rename")?;
    }

    index::insert(new_name, value);

    Ok(())
}

pub fn update_payload_data(name: impl AsRef<str>, data: impl AsRef<[u8]>) -> Result<()> {
    let name = name.as_ref();
    let external = index::get(name).map(|v| v.external).unwrap_or(false);

    if external {
        index::remove(name);
    } else if exists_in_index(&name) {
        remove(&name).context("upd")?;
    }

    put(name, data).context("upd_put")
}

/// Add an existing file to index
pub fn put_external(path: impl AsRef<Path>) -> Result<()> {
    let path = path.as_ref();
    let path_str = path.as_os_str().to_string_lossy().to_string();

    if exists_in_index(&path_str) {
        return Ok(());
    }

    if !path.exists() || path.is_dir() {
        return Err(Error::not_found(path)).context("ext1");
    }

    let file_data = fs::read(&path_str).unwrap();
    let data_hash = hash_data(&file_data);

    index::insert(
        path_str,
        IndexFileMeta::new(path, data_hash, true).context("ext2")?,
    );

    Ok(())
}

pub fn stats() -> Vec<IndexFileMeta> {
    index::stats()
}

/*
pub fn cleanup() {
    let index = INDEX.read();

    index.values().for_each(|f| {
        fs::remove_file(&f.path).ok();
    });
}
*/

fn hash_data(data: impl AsRef<[u8]>) -> String {
    use sha3::{Digest, Sha3_256};

    let mut hasher = Sha3_256::new();
    hasher.update(data);
    let hashsum = hasher.finalize().to_vec();
    hex::encode(hashsum)
}
