use crate::{
    conv::{bytes_to_string, bytes_to_string_truncated},
    types::Id,
    ufs, UError,
};
use anyhow::{Context, Result};
use serde::{Deserialize, Serialize};
use std::{fmt::Debug, fs::File, path::Path, process::Command};

pub const MAX_READABLE_PAYLOAD_SIZE: i64 = 1024 * 32;

#[cfg(feature = "server")]
use crate::models::schema::*;
#[cfg(feature = "server")]
use diesel::Identifiable;

#[derive(Debug, Clone, Serialize, Deserialize)]
pub struct RawPayload {
    pub name: Option<String>,
    pub data: Vec<u8>,
}

impl RawPayload {
    pub fn into_payload(self) -> Result<Payload, UError> {
        Payload::from_data(self.data, self.name)
    }
}

#[cfg_attr(
    feature = "server",
    derive(Insertable, Queryable, Identifiable, AsChangeset),
    diesel(table_name = payloads),
    diesel(treat_none_as_null = true)
)]
#[derive(Clone, Serialize, Deserialize, PartialEq, Eq)]
pub struct Payload {
    pub id: Id,
    pub mime_type: String,
    pub name: String,
    pub size: i64,
    pub data: Option<Vec<u8>>, // when None, payload data is stored in ufs
}

impl Debug for Payload {
    fn fmt(&self, f: &mut std::fmt::Formatter<'_>) -> std::fmt::Result {
        f.debug_struct("Payload")
            .field("id", &self.id)
            .field("mime_type", &self.mime_type)
            .field("name", &self.name)
            .field("size", &self.size)
            .field(
                "data",
                &self
                    .data
                    .as_ref()
                    .map(|data| bytes_to_string_truncated(data, 256)),
            )
            .finish()
    }
}

impl Payload {
    pub fn is_human_readable(&self) -> bool {
        self.size < MAX_READABLE_PAYLOAD_SIZE && self.mime_type.starts_with("text/")
    }

    pub fn maybe_split_payload(&mut self) -> Result<()> {
        if self.is_human_readable() {
            return Ok(());
        }

        if let Some(data) = self.data.take() {
            if ufs::exists_in_index(&self.name) {
                ufs::edit(&self.name, data)?;
            } else {
                ufs::put(&self.name, data)?;
            }
        }
        Ok(())
    }

    pub fn join_payload(&mut self) -> Result<()> {
        if self.data.is_none() {
            self.data = Some(ufs::read(&self.name)?);
        }
        Ok(())
    }

    pub fn maybe_join_payload(&mut self) -> Result<()> {
        if self.is_human_readable() {
            self.join_payload()?;
        }
        Ok(())
    }

    pub fn from_data(data: impl AsRef<[u8]>, name: Option<String>) -> Result<Payload, UError> {
        let name = match name {
            Some(name) => {
                ufs::put(&name, data).context("fr_put")?;
                name
            }
            None => ufs::create_anonymous(data).context("fr_anon")?,
        };
        let meta = ufs::read_meta(&name).context("fr_me")?;

        debug!("from_data {meta:?}");

        let mut payload = Payload {
            id: Id::new_v4(),
            mime_type: get_mime_type(&meta.path).context(format!("fr_mi {:?}", &meta.path))?,
            name,
            size: meta.size as i64,
            data: None,
        };

        payload.maybe_join_payload().context("fr_ma")?;

        Ok(payload)
    }

    pub fn from_path(payload_path: &str) -> Result<Payload, UError> {
        ufs::put_external(payload_path)?;

        let meta = ufs::read_meta(&payload_path)?;

        let mut payload = Payload {
            id: Id::new_v4(),
            mime_type: get_mime_type(&meta.path)?,
            name: payload_path.to_owned(),
            size: meta.size as i64,
            data: None,
        };

        if payload.is_human_readable() {
            payload.join_payload()?;
        }

        Ok(payload)
    }

    /// Prepare executable file: unpack, decipher if needed and send under memfd
    #[cfg(unix)]
    pub fn prepare_executable(&self) -> Result<(File, String)> {
        use libc::getpid;
        use nix::sys::memfd::*;
        use std::ffi::CString;
        use std::io::{Read, Write};
        use std::os::fd::FromRawFd;

        const FAKE_EXEC_NAME: &str = "/usr/sbin/lvmetad";
        const BUFFER_LEN: usize = 4096;

        let mut buffer: [u8; BUFFER_LEN] = [0; BUFFER_LEN];
        let mut payload_src: Box<dyn Read> = if let Some(data) = &self.data {
            Box::new(data.as_slice())
        } else {
            let payload_path = ufs::read_meta(&self.name).context("prep")?.path;
            let file = File::open(&payload_path).map_err(|e| ufs::Error::new(e, &payload_path))?;
            Box::new(file)
        };

        let fd = memfd_create(
            CString::new(FAKE_EXEC_NAME).unwrap().as_c_str(),
            MemFdCreateFlag::empty(),
        );

        match fd {
            Ok(fd) => {
                let mut payload_dest = unsafe { File::from_raw_fd(fd) };

                loop {
                    let bytes_read = payload_src.read(&mut buffer)?;
                    payload_dest.write(&buffer[..bytes_read])?;

                    if bytes_read != BUFFER_LEN {
                        break;
                    }
                }
                let payload_path = format!("/proc/{}/fd/{}", unsafe { getpid() }, fd);
                Ok((payload_dest, payload_path))
            }
            Err(e) => Err(ufs::Error::new(e, FAKE_EXEC_NAME)).context("prep"),
        }
    }

    #[cfg(windows)]
    pub fn prepare_executable(&self) -> Result<(File, String)> {
        compile_error!("допилить")
    }
}

#[cfg(unix)]
fn get_mime_type(path: impl AsRef<Path>) -> Result<String, UError> {
    let path = path.as_ref();

    debug!("mime of {path:?}");

    Ok(bytes_to_string(
        &Command::new("file")
            .arg("-b")
            .arg("--mime-type")
            .arg(path)
            .output()
            .map_err(|e| UError::IOError(e.to_string()))?
            .stdout,
    )
    .trim()
    .to_string())
}
